# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test package contains the tests for ipfs registry tools."""

from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

import pytest
from aea_cli_ipfs.exceptions import HashNotProvided
from aea_cli_ipfs.registry import (
    LOCAL_REGISTRY_DEFAULT,
    fetch_ipfs,
    get_ipfs_hash_from_public_id,
    load_local_registry,
    register_item_to_local_registry,
    validate_registry,
    write_local_registry,
)

from aea.configurations.data_types import PublicId


DUMMY_REGISTRY_DATA = LOCAL_REGISTRY_DEFAULT.copy()
DUMMY_REGISTRY_DATA["protocols"] = {
    "default_author/component:0.1.0": "hash_1",
    "default_author/component:0.2.0": "hash_2",
    "default_author/component:0.3.0": "hash_3",
}


def test_validate_registry() -> None:
    """Test validate registry."""

    with pytest.raises(ValueError):
        validate_registry({})


def test_write_local_registry() -> None:
    """Test write_local_registry method."""

    with TemporaryDirectory() as temp_dir:
        registry_path = Path(temp_dir) / "local_registry.json"
        write_local_registry(LOCAL_REGISTRY_DEFAULT, str(registry_path))

        assert registry_path.is_file()
        assert load_local_registry(str(registry_path)) == LOCAL_REGISTRY_DEFAULT


def test_load_local_registry() -> None:
    """Test write_local_registry method."""

    with TemporaryDirectory() as temp_dir:
        registry_path = Path(temp_dir) / "local_registry.json"
        assert load_local_registry(str(registry_path)) == LOCAL_REGISTRY_DEFAULT

        write_local_registry(DUMMY_REGISTRY_DATA, str(registry_path))
        assert load_local_registry(str(registry_path)) == DUMMY_REGISTRY_DATA


def test_get_ipfs_hash_from_public_id() -> None:
    """Test get_ipfs_hash_from_public_id method."""

    with mock.patch(
        "aea_cli_ipfs.registry.load_local_registry",
        new=lambda *_, **__: DUMMY_REGISTRY_DATA,
    ):
        # test hash retrival
        package_hash = get_ipfs_hash_from_public_id(
            "protocol", PublicId.from_str("default_author/component:0.1.0")
        )
        assert package_hash == "hash_1"

        # test hash retrival for latest package
        package_hash = get_ipfs_hash_from_public_id(
            "protocol", PublicId.from_str("default_author/component:latest")
        )
        assert package_hash == "hash_3"

        # test hash retrival for non existing package
        package_hash = get_ipfs_hash_from_public_id(
            "protocol", PublicId.from_str("default_author/component:1.0.0")
        )
        assert package_hash is None

        # test hash retrival for latest package not exists
        package_hash = get_ipfs_hash_from_public_id(
            "protocol", PublicId.from_str("default_author/component1:latest")
        )
        assert package_hash is None


def test_register_item_to_local_registry() -> None:
    """Test register_item_to_local_registry method."""

    with TemporaryDirectory() as temp_dir:
        registry_path = Path(temp_dir) / "local_registry.json"
        write_local_registry(DUMMY_REGISTRY_DATA, str(registry_path))
        register_item_to_local_registry(
            "skill", "fetchai/skill:0.1.0", "skill_hash", registry_path=registry_path
        )
        assert (
            get_ipfs_hash_from_public_id(
                "skill",
                PublicId.from_str("fetchai/skill:0.1.0"),
                registry_path=registry_path,
            )
            == "skill_hash"
        )

        assert (
            get_ipfs_hash_from_public_id(
                "skill",
                PublicId.from_str("fetchai/skill:latest"),
                registry_path=registry_path,
            )
            == "skill_hash"
        )


def test_fetch_ipfs() -> None:
    """Test fetch_ipfs method."""

    with mock.patch(
        "aea_cli_ipfs.registry.load_local_registry",
        new=lambda *_, **__: DUMMY_REGISTRY_DATA,
    ), TemporaryDirectory() as dest_path:
        with mock.patch("aea_cli_ipfs.ipfs_utils.IPFSTool.download"), mock.patch(
            "aea_cli_ipfs.ipfs_utils.IPFSTool.check_ipfs_node_running"
        ):
            package_path = fetch_ipfs(
                "protocol",
                PublicId.from_str(
                    "default_author/component:0.2.0:QmYAXgX8ARiriupMQsbGXtKdDyGzWry1YV3sycKw1qqmgH"
                ),
                dest_path,
            )
            assert package_path == Path(dest_path).absolute()

            package_path = fetch_ipfs(
                "protocol",
                PublicId.from_str(
                    "default_author/component:0.2.0:QmYAXgX8ARiriupMQsbGXtKdDyGzWry1YV3sycKw1qqmgH"
                ),
                dest_path,
                remote=False,
            )
            assert package_path == Path(dest_path).absolute()

            # no public_id hash
            package_path = fetch_ipfs(
                "protocol",
                PublicId.from_str("default_author/component:0.2.0"),
                dest_path,
                remote=False,
            )
            assert package_path == Path(dest_path).absolute()

            # no public_id hash  remote=True
            with pytest.raises(HashNotProvided):
                fetch_ipfs(
                    "protocol",
                    PublicId.from_str("default_author/component:0.2.0"),
                    dest_path,
                    remote=True,
                )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test ipfs utils."""

import os
import re
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import Mock, patch

import ipfshttpclient
import pytest
import requests
from aea_cli_ipfs.ipfs_utils import (
    DownloadError,
    IPFSDaemon,
    IPFSTool,
    PinError,
    RemoveError,
    resolve_addr,
)
from aea_cli_ipfs.test_tools.fixture_helpers import ipfs_daemon  # noqa: F401

from aea.cli.registry.settings import DEFAULT_IPFS_URL, DEFAULT_IPFS_URL_LOCAL


def test_init_tool() -> None:
    """Test tool initialization."""

    tool = IPFSTool(DEFAULT_IPFS_URL)
    assert tool.is_remote is True

    tool = IPFSTool(DEFAULT_IPFS_URL_LOCAL)
    assert tool.is_remote is False


def test_hash_bytes() -> None:
    """Test hash bytes."""
    tool = IPFSTool(DEFAULT_IPFS_URL_LOCAL)
    tool.daemon.start()
    try:
        assert tool.is_remote is False
        some_bytes = b"there is some bytes"
        ipfs_hash = tool.add_bytes(some_bytes)
        assert (
            ipfs_hash == "QmPPFcK8uynDmceTDkDjHuDbR7gnqBfaMCTifs6oFKHn4E"
        )  # precalculated
    finally:
        tool.daemon.stop()


def test_resolve_addr() -> None:
    """Test resolve_addr function."""
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Address type should be one of the ('ip4', 'dns'), provided: 2"
        ),
    ):
        resolve_addr("1/2/3/4/5/6")

    with pytest.raises(
        ValueError,
        match=re.escape("Connection should be one of the ('tcp',), provided: 4"),
    ):
        resolve_addr("1/ip4/3/4/5/6")

    with pytest.raises(
        ValueError,
        match=re.escape("Protocol should be one of the ('http', 'https'), provided: 6"),
    ):
        resolve_addr("1/ip4/3/tcp/5/6")

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Invalid multiaddr string provided, valid format: /{dns,dns4,dns6,ip4}/<host>/tcp/<port>/protocol. Provided: 1/ip4/3/tcp/5/https/1/2"
        ),
    ):
        resolve_addr("1/ip4/3/tcp/5/https/1/2")


def test_daemon_url_trim() -> None:
    """Test extra slash in addr url trimmed."""
    daemon = IPFSDaemon(is_remote=False, node_url="http://1.1.1.1:1/")
    assert daemon.node_url == "http://1.1.1.1:1"


def test_daemon_ipfs_not_found() -> None:
    """Test ipfs cli command avialable."""
    with patch("shutil.which", return_value=None):
        with pytest.raises(Exception, match="Please install IPFS first!"):
            IPFSDaemon._check_ipfs()

    with patch("subprocess.Popen.communicate", return_value=(b"", b"")):
        with pytest.raises(
            Exception,
            match=re.escape(
                "Please ensure you have version 0.6.0 of IPFS daemon installed."
            ),
        ):
            IPFSDaemon._check_ipfs()


def test_daemon_is_started_externally() -> None:
    """Test IPFSDaemon is started externally."""
    daemon = IPFSDaemon()
    response_mock = Mock()
    response_mock.status_code = 200
    with patch("requests.post", return_value=response_mock):
        assert daemon.is_started_externally()

    response_mock.status_code = 400
    with patch("requests.post", return_value=response_mock):
        assert not daemon.is_started_externally()

    with patch("requests.post", side_effect=requests.exceptions.ConnectionError()):
        assert not daemon.is_started_externally()

    assert not daemon.is_started()


def test_daemon_start() -> None:
    """Test IPFSDaemon start method."""
    daemon = IPFSDaemon()
    popen_mock = Mock()
    popen_mock.communicate = Mock(return_value=(b"0.6.0", b""))
    popen_mock.stdout = None
    with patch("subprocess.Popen", return_value=popen_mock):
        with pytest.raises(RuntimeError, match="Could not start IPFS daemon."):
            daemon.start()

        popen_mock.stdout = Mock()
        popen_mock.stdout.readline = Mock(side_effect=[b""] * 10)
        with pytest.raises(RuntimeError, match="Could not start IPFS daemon."):
            daemon.start()


def test_daemon_context() -> None:
    """Test IPFSDaemon context manager."""
    daemon = IPFSDaemon()
    with patch.object(daemon, "start") as start_mock, patch.object(
        daemon, "stop"
    ) as stop_mock:
        with daemon:
            pass

        start_mock.assert_called_once()
        stop_mock.assert_called_once()


def test_tool_all_pins() -> None:
    """Test IPFSTool.all_pins and is_a_package method."""
    ipfs_tool = IPFSTool()
    client_mock = Mock()
    client_mock.pin.ls = Mock(return_value={"Keys": [1, 2, 3, 4, 5]})
    with patch.object(ipfs_tool, "client", client_mock):
        assert ipfs_tool.all_pins() == {1, 2, 3, 4, 5}
        assert ipfs_tool.is_a_package(1)
        assert not ipfs_tool.is_a_package(6)


def test_tool_add_pin() -> None:
    """Test IPFSTool.pin method."""
    ipfs_tool = IPFSTool()
    client_mock = Mock()
    client_mock.pin.add = Mock(return_value={"Keys": [1]})
    with patch.object(ipfs_tool, "client", client_mock):
        assert ipfs_tool.pin("some")

        with pytest.raises(PinError):
            client_mock.pin.add = Mock(
                side_effect=ipfshttpclient.exceptions.ErrorResponse(Mock(), Mock())
            )
            ipfs_tool.pin("some")


def test_tool_remove_unpinned_files() -> None:
    """Test IPFSTool.remove_unpinned_files method."""
    ipfs_tool = IPFSTool()
    client_mock = Mock()
    client_mock.repo.gc = Mock(return_value={"Keys": [1]})
    with patch.object(ipfs_tool, "client", client_mock):
        ipfs_tool.remove_unpinned_files()

        with pytest.raises(RemoveError):
            client_mock.repo.gc = Mock(
                side_effect=ipfshttpclient.exceptions.ErrorResponse(Mock(), Mock())
            )
            ipfs_tool.remove_unpinned_files()


def test_tool_download() -> None:
    """Test IPFSTool.download method."""
    ipfs_tool = IPFSTool()
    client_mock = Mock()
    client_mock.get = Mock(
        side_effect=ipfshttpclient.exceptions.StatusError(Mock(), Mock())
    )
    with patch.object(
        ipfs_tool, "client", client_mock
    ), TemporaryDirectory() as tmp_dir, patch(
        "pathlib.Path.is_file", return_value=True
    ), patch(
        "shutil.copy"
    ), patch(
        "os.listdir", return_value=["1"]
    ), patch(
        "pathlib.Path.iterdir", return_value=[]
    ), patch(
        "shutil.rmtree"
    ), patch(
        "shutil.move"
    ):
        with pytest.raises(DownloadError, match="Failed to download: some"):
            with patch("time.sleep"):
                ipfs_tool.download("some", tmp_dir, attempts=5)
        assert client_mock.get.call_count == 5

        client_mock.get = Mock()

        assert ipfs_tool.download("some", tmp_dir, attempts=5) == tmp_dir


def test_tool_download_fix_path_works() -> None:
    """Test IPFSTool.download method."""
    ipfs_tool = IPFSTool()
    client_mock = Mock()
    sub_file_name = "some1"
    hash_id = "some_hash"

    def make_files(hash_, download_dir):
        os.mkdir(Path(download_dir) / hash_)
        os.mkdir(Path(download_dir) / hash_ / sub_file_name)

    client_mock.get = make_files
    with patch.object(ipfs_tool, "client", client_mock), patch(
        "pathlib.Path.is_file", return_value=False
    ), patch("shutil.copy"), patch("os.listdir", return_value=["1"]), patch(
        "pathlib.Path.iterdir", return_value=[]
    ), patch(
        "shutil.rmtree"
    ):
        with TemporaryDirectory() as target_tmp_dir:
            assert ipfs_tool.download(hash_id, target_tmp_dir, fix_path=True) == str(
                Path(target_tmp_dir, sub_file_name)
            )
            assert [i.name for i in Path(target_tmp_dir).glob("*")] == [sub_file_name]

        with TemporaryDirectory() as target_tmp_dir:
            assert ipfs_tool.download(hash_id, target_tmp_dir, fix_path=False) == str(
                Path(target_tmp_dir) / hash_id
            )
            assert [i.name for i in Path(target_tmp_dir).glob("*")] == [hash_id]


def test_wrap_directory_flag_file() -> None:
    """Test `wrap_directory` flag"""

    tool = IPFSTool(DEFAULT_IPFS_URL_LOCAL)
    tool.daemon.start()
    try:
        with TemporaryDirectory() as temp_dir:
            temp_file = Path(temp_dir, "txt")
            temp_file.write_text("Hello, World")

            _, file_hash, _ = tool.add(dir_path=str(temp_file))
            assert file_hash == "QmWVQQhQ5Qxzb1jLk1SW4Etsn6rMWHtjdELTNEmA1J1gRx"

            _, file_hash, _ = tool.add(
                dir_path=str(temp_file), wrap_with_directory=False
            )
            assert file_hash == "QmTev1ZgJkHgFYiCX7MgELEDJuMygPNGcinqBa2RmfnGFu"
    finally:
        tool.daemon.stop()


def test_wrap_directory_flag_dir() -> None:
    """Test `wrap_directory` flag"""
    tool = IPFSTool(DEFAULT_IPFS_URL_LOCAL)
    tool.daemon.start()
    try:
        with TemporaryDirectory() as _temp_dir:
            temp_dir = Path(_temp_dir, "some_dir")
            temp_dir.mkdir()

            temp_file = temp_dir / "txt"
            temp_file.write_text("Hello, World")

            _, file_hash, _ = tool.add(dir_path=str(temp_dir))
            assert file_hash == "Qmb7LSaArLheRjnvVZ2vhhnBun8EWsF9Z5TdL8NgLgyhJL"

            _, file_hash, _ = tool.add(
                dir_path=str(temp_dir), wrap_with_directory=False
            )
            assert file_hash == "QmWVQQhQ5Qxzb1jLk1SW4Etsn6rMWHtjdELTNEmA1J1gRx"
    finally:
        tool.daemon.stop()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Tests for aea cli ipfs plugin."""

import os
import re
import sys
from pathlib import Path
from unittest import mock
from unittest.mock import patch

import click
import ipfshttpclient  # type: ignore
import pytest
from aea_cli_ipfs.ipfs_utils import addr_to_url, resolve_addr
from click.testing import CliRunner
from urllib3.exceptions import NewConnectionError as ConnectionError

from aea.cli.core import cli
from aea.test_tools.click_testing import CliTest


sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

from aea_cli_ipfs.core import (  # noqa # type: ignore  # pylint: disable=wrong-import-position
    PublishError,
    ipfs,
)


DUMMY_HASH = "QmbWqxBEKC3P8tqsKc98xmWNzrzDtRLMiMPL8wBuTGsMnR"

cli.add_command(ipfs)


def test_addr_helpers():
    """Test `resolve_addr` method."""

    addr_scheme, host, conn_type, port, protocol = resolve_addr(
        "/ip4/127.0.0.1/tcp/5001/http"
    )

    assert addr_scheme == "ip4"
    assert host == "127.0.0.1"
    assert conn_type == "tcp"
    assert port == "5001"
    assert protocol == "http"

    *_, port, protocol = resolve_addr("/ip4/127.0.0.1/tcp")

    assert port == "5001"
    assert protocol == "http"

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Address type should be one of the ('ip4', 'dns'), provided: ip6"
        ),
    ):
        resolve_addr("/ip6/127.0.0.1/tcp")

    with pytest.raises(
        ValueError,
        match=re.escape("Connection should be one of the ('tcp',), provided: udp"),
    ):
        resolve_addr("/ip4/127.0.0.1/udp")

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Protocol should be one of the ('http', 'https'), provided: wss"
        ),
    ):
        resolve_addr("/ip4/127.0.0.1/tcp/5001/wss")

    assert addr_to_url("/ip4/127.0.0.1/tcp/5001/http") == "http://127.0.0.1:5001"


def test_ipfs():
    """Test aea ipfs command itself."""
    runner = CliRunner()
    with patch("ipfshttpclient.Client.id"):
        r = runner.invoke(cli, ["ipfs"], catch_exceptions=False, standalone_mode=False)
    assert r.exit_code == 0


def test_ipfs_add():
    """Test aea ipfs add."""
    runner = CliRunner()
    with patch("ipfshttpclient.Client.name.publish") as ipfs_publish, patch(
        "ipfshttpclient.Client.id"
    ) as ipfs_id, patch(
        "ipfshttpclient.Client.add",
        return_value=[{"Name": "name", "Hash": DUMMY_HASH}] * 2,
    ) as ipfs_add, patch(
        "aea_cli_ipfs.ipfs_utils.IPFSDaemon._check_ipfs", new=lambda *_: None
    ):
        r = runner.invoke(cli, ["ipfs", "add", "-p"], catch_exceptions=False)
    assert r.exit_code == 0
    ipfs_id.assert_called()
    ipfs_add.assert_called()
    ipfs_publish.assert_called()

    with patch(
        "ipfshttpclient.Client.name.publish", side_effect=PublishError("oops")
    ) as ipfs_publish, patch("ipfshttpclient.Client.id") as ipfs_id, patch(
        "ipfshttpclient.Client.add",
        return_value=[{"Name": "name", "Hash": DUMMY_HASH}] * 2,
    ) as ipfs_add, patch(
        "aea_cli_ipfs.ipfs_utils.IPFSDaemon._check_ipfs", new=lambda *_: None
    ):
        with pytest.raises(click.ClickException, match="Publish failed.*oops"):
            runner.invoke(
                cli,
                ["ipfs", "add", "-p"],
                catch_exceptions=False,
                standalone_mode=False,
            )


def test_node_not_alive_can_not_be_started():
    """Test error on node connection failed"""
    runner = CliRunner()
    with patch(
        "ipfshttpclient.Client.id",
        side_effect=ipfshttpclient.exceptions.CommunicationError(
            original=Exception("oops")
        ),
    ), patch("time.sleep"), patch("subprocess.Popen"), patch(
        "aea_cli_ipfs.ipfs_utils.IPFSDaemon._check_ipfs"
    ), patch(
        "aea_cli_ipfs.ipfs_utils.IPFSDaemon.start"
    ), patch(
        "aea_cli_ipfs.ipfs_utils.IPFSDaemon._check_ipfs", new=lambda *_: None
    ):
        env = os.environ.copy()
        env["OPEN_AEA_IPFS_ADDR"] = "/ip4/127.0.0.1/tcp/5001"
        with pytest.raises(ConnectionError):
            runner.invoke(
                cli,
                ["ipfs", "add", "-p"],
                catch_exceptions=False,
                standalone_mode=False,
                env=env,
            )


@pytest.mark.skip
def test_version_did_not_match():
    """Test error on node connection failed"""
    runner = CliRunner()
    with patch(
        "ipfshttpclient.Client.id",
        side_effect=ipfshttpclient.exceptions.CommunicationError(
            original=Exception("oops")
        ),
    ), patch("time.sleep"), patch(
        "subprocess.Popen.communicate", new_callable=lambda: lambda _: (b"", None)
    ), patch(
        "aea_cli_ipfs.ipfs_utils.IPFSDaemon._check_ipfs", new=lambda *_: None
    ):
        with pytest.raises(
            Exception,
            match="Please ensure you have version 0.6.0 of IPFS daemon installed.",
        ):
            runner.invoke(
                cli,
                ["ipfs", "add", "-p"],
                catch_exceptions=False,
                standalone_mode=False,
            )


class TestIPFSToolDownload(CliTest):
    """Test IPFSTool.download"""

    # we download either a file or a directory
    # if a file: the original name is not preserved, and will be the hash
    #   e.g. <filename> --> <some_ipfs_hash>
    # if a dir : the original name is preserved, as a nested directory
    #   e.g. <directory>/ --> <some_ipfs_hash>/<directory>/

    cli_options = ("ipfs", "download")

    def setup(self) -> None:
        """Setup"""

        super().setup()
        self.some_ipfs_hash = "not_a_real_ipfs_hash"
        self.target_dir = self.t / "target_dir"
        self.args = self.some_ipfs_hash, str(self.target_dir)

    @staticmethod
    def mock_client_get_success(is_dir: bool) -> mock._patch:
        """Mock IPFSTool.client.get"""

        # self.client.get(hash_id, tmp_dir) creates tmp_dir/hash_id

        def new_callable(_, hash_id, tmp_dir) -> None:
            if is_dir:  # <ipfs_hash>/<dir_name>/...
                some_dir = Path(tmp_dir) / hash_id / "some_dir"
                some_dir.mkdir(parents=True)
                (some_dir / "some_file").touch()
                (some_dir / "some_nested_dir").mkdir()
                (some_dir / "some_nested_dir" / "some_nested_file").touch()
            else:  # <ipfs_hash>
                (Path(tmp_dir) / hash_id).touch()

        # we need a nested lambda to mock a method on the class instead of instance
        return patch("ipfshttpclient.Client.get", new_callable=lambda: new_callable)

    @property
    def mock_client_get_failure(self) -> mock._patch:
        """Mock IPFSTool.client.get failure"""

        def new_callable(*_, **__) -> None:
            exception = Exception("DummyError for testing")
            raise ipfshttpclient.exceptions.StatusError(exception)

        return patch("ipfshttpclient.Client.get", new_callable=lambda: new_callable)

    @pytest.mark.parametrize("is_dir", [False, True])
    def test_ipfs_download_success(self, is_dir: bool) -> None:
        """Test aea ipfs download."""

        with self.mock_client_get_success(is_dir=is_dir):
            result = self.run_cli(*self.args, catch_exceptions=False)

        assert result.exit_code == 0, result.stdout
        assert f"Download {self.some_ipfs_hash} to {self.target_dir}" in result.stdout
        assert "Download complete!" in result.stdout

        all_new_paths = list(self.target_dir.rglob("*"))

        if is_dir:
            assert len(all_new_paths) == 4
            assert not any(self.some_ipfs_hash in str(p) for p in all_new_paths)
        else:
            assert len(all_new_paths) == 1
            assert all_new_paths[0].is_file()
            assert all_new_paths[0].name == self.some_ipfs_hash

    def test_ipfs_download_failure(self) -> None:
        """Test aea ipfs download failure."""

        with self.mock_client_get_failure:
            with mock.patch("time.sleep"):
                result = self.run_cli(
                    *self.args, catch_exceptions=True, standalone_mode=False
                )

        assert result.exit_code == 1, result.stdout
        assert isinstance(result.exception, click.ClickException)
        assert f"Failed to download: {self.some_ipfs_hash}" in result.exception.message
        assert not any(self.target_dir.rglob("*"))


@patch("ipfshttpclient.Client.id")
def test_ipfs_remove(*_):
    """Test aea ipfs remove."""
    runner = CliRunner()
    with patch("ipfshttpclient.Client.pin.rm") as ipfs_rm, patch(
        "aea_cli_ipfs.ipfs_utils.IPFSDaemon._check_ipfs", new=lambda *_: None
    ):
        r = runner.invoke(cli, ["ipfs", "remove", "some_hash"], catch_exceptions=False)
    assert r.exit_code == 0
    ipfs_rm.assert_called()

    with patch(
        "ipfshttpclient.Client.pin.rm",
        side_effect=ipfshttpclient.exceptions.ErrorResponse(
            "oops", original=Exception()
        ),
    ) as ipfs_rm, patch(
        "aea_cli_ipfs.ipfs_utils.IPFSDaemon._check_ipfs", new=lambda *_: None
    ):
        with pytest.raises(click.ClickException, match="Remove error:.*oops"):
            runner.invoke(
                cli,
                ["ipfs", "remove", "some_hash"],
                catch_exceptions=False,
                standalone_mode=False,
            )


if __name__ == "__main__":
    pytest.main([__file__])



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Ipfs utils for `ipfs cli command`."""

import logging
import os
import shutil
import signal
import subprocess  # nosec
import tempfile
import time
from pathlib import Path
from typing import Dict, IO, List, Optional, Set, Tuple, Union, cast

import ipfshttpclient  # type: ignore
import requests
from aea_cli_ipfs.exceptions import (
    DownloadError,
    NodeError,
    PinError,
    PublishError,
    RemoveError,
)


DEFAULT_IPFS_URI_BASE = str(os.environ.get("OPEN_AEA_IPFS_ADDR_BASE", "api/v0"))
DEFAULT_IPFS_URL = "/dns/registry.autonolas.tech/tcp/443/https"
DEFAULT_IPFS_URL_LOCAL = "/ip4/127.0.0.1/tcp/5001"
ALLOWED_CONNECTION_TYPES = ("tcp",)
ALLOWED_ADDR_TYPES = ("ip4", "dns")
ALLOWED_PROTOCOL_TYPES = ("http", "https")
MULTIADDR_FORMAT = "/{dns,dns4,dns6,ip4}/<host>/tcp/<port>/protocol"
IPFS_NODE_CHECK_ENDPOINT = "/api/v0/id"
IPFS_VERSION = "0.6.0"


def _verify_attr(name: str, attr: str, allowed: Tuple[str, ...]) -> None:
    """Varify various attributes of ipfs address."""

    if attr not in allowed:
        raise ValueError(f"{name} should be one of the {allowed}, provided: {attr}")


def resolve_addr(addr: str) -> Tuple[str, ...]:
    """
    Multiaddr resolver.

    :param addr: multiaddr string.
    :return: http URL
    """
    _, addr_scheme, host, conn_type, *extra_data = addr.split("/")

    if len(extra_data) > 2:  # pylint: disable=no-else-raise
        raise ValueError(
            f"Invalid multiaddr string provided, valid format: {MULTIADDR_FORMAT}. Provided: {addr}"
        )
    elif len(extra_data) == 2:
        port, protocol, *_ = extra_data
    elif len(extra_data) == 1:
        (port,) = extra_data
        protocol = "http"
    else:
        port = "5001"
        protocol = "http"

    _verify_attr("Address type", addr_scheme, ALLOWED_ADDR_TYPES)
    _verify_attr("Connection", conn_type, ALLOWED_CONNECTION_TYPES)
    _verify_attr("Protocol", protocol, ALLOWED_PROTOCOL_TYPES)

    return addr_scheme, host, conn_type, port, protocol


def addr_to_url(addr: str) -> str:
    """Convert address to url."""

    _, host, _, port, protocol = resolve_addr(addr)
    return f"{protocol}://{host}:{port}"


def is_remote_addr(host: str) -> bool:
    """Check if addr is remote or local."""
    return host not in ("localhost", "127.0.0.1", "0.0.0.0")  # nosec


class IPFSDaemon:
    """
    Set up the IPFS daemon.

    :raises Exception: if IPFS is not installed.
    """

    api_url: str
    node_url: str
    is_remote: bool
    process: Optional[subprocess.Popen]

    def __init__(
        self, node_url: str = "http://127.0.0.1:5001", is_remote: bool = False
    ):
        """Initialise IPFS daemon."""

        if node_url.endswith("/"):
            node_url = node_url[:-1]

        self.node_url = node_url
        self.api_url = node_url + IPFS_NODE_CHECK_ENDPOINT
        self.process = None
        self.is_remote = is_remote

        if not is_remote:
            self._check_ipfs()

    @staticmethod
    def _check_ipfs() -> None:
        """Check if IPFS node is running."""
        res = shutil.which("ipfs")
        if res is None:
            raise Exception("Please install IPFS first!")
        process = subprocess.Popen(  # nosec
            ["ipfs", "--version"],
            stdout=subprocess.PIPE,
            env=os.environ.copy(),
        )
        output, _ = process.communicate()
        if b"0.6.0" not in output:
            raise Exception(
                "Please ensure you have version 0.6.0 of IPFS daemon installed."
            )

    def is_started_externally(self) -> bool:
        """Check daemon was started externally."""
        try:
            x = requests.post(self.api_url)
            return x.status_code == 200
        except requests.exceptions.ConnectionError:
            return False

    def is_started_internally(self) -> bool:
        """Check daemon was started internally."""
        return bool(self.process)

    def is_started(self) -> bool:
        """Check daemon was started."""
        return self.is_started_externally() or self.is_started_internally()

    def start(self) -> None:
        """Run the ipfs daemon."""
        cmd = ["ipfs", "daemon", "--offline"]
        self.process = subprocess.Popen(  # nosec
            cmd,
            stdout=subprocess.PIPE,
            env=os.environ.copy(),
        )
        empty_outputs = 0

        if self.process.stdout is None:
            raise RuntimeError("Could not start IPFS daemon.")

        for stdout_line in iter(cast(IO[bytes], self.process.stdout).readline, ""):
            if b"Daemon is ready" in stdout_line:
                break
            if stdout_line == b"":
                empty_outputs += 1
                if empty_outputs >= 5:
                    raise RuntimeError("Could not start IPFS daemon.")

    def stop(self) -> None:  # pragma: nocover
        """Terminate the ipfs daemon if it was started internally."""
        if self.process is None:
            return

        if self.process.stdout is not None:
            self.process.stdout.close()

        self.process.send_signal(signal.SIGTERM)
        self.process.wait(timeout=30)
        poll = self.process.poll()
        if poll is None:
            self.process.terminate()
            self.process.wait(2)
        self.process = None

    def __enter__(self) -> None:
        """Run the ipfs daemon."""
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        """Terminate the ipfs daemon."""
        self.stop()


class IPFSTool:
    """IPFS tool to add, publish, remove, download directories."""

    _addr: Optional[str] = None

    def __init__(
        self,
        addr: Optional[str] = None,
        base: str = DEFAULT_IPFS_URI_BASE,
    ):
        """
        Init tool.

        :param addr: multiaddr string for IPFS client.
        :param base: API base for IPFS client.
        """

        if addr is None:
            addr = os.environ.get("OPEN_AEA_IPFS_ADDR", DEFAULT_IPFS_URL)

        _, host, *_ = resolve_addr(cast(str, addr))  # verify addr

        self._addr = addr
        self.is_remote = is_remote_addr(host)
        self.client = ipfshttpclient.Client(addr=self.addr, base=base)
        self.daemon = IPFSDaemon(
            node_url=addr_to_url(self.addr), is_remote=self.is_remote
        )

    @property
    def addr(
        self,
    ) -> str:
        """Node address"""
        return cast(str, self._addr)

    def is_a_package(self, package_hash: str) -> bool:
        """Checks if a package with `package_hash` is pinned or not"""
        return package_hash in self.all_pins()

    def all_pins(self, recursive_only: bool = True) -> Set[str]:
        """Returns a list of all pins."""
        pinned_hashes = self.client.pin.ls(
            type="recursive" if recursive_only else "all"
        )
        return set(pinned_hashes["Keys"])

    def add_bytes(self, data: bytes, **kwargs) -> str:
        """
        Add bytes data to ipfs.

        :param data: bytes
        :param kwargs: options passed to request library

        :return: hash
        """
        response = self.client.add_bytes(data, **kwargs)
        return response

    def add(
        self,
        dir_path: str,
        pin: bool = True,
        recursive: bool = True,
        wrap_with_directory: bool = True,
    ) -> Tuple[str, str, List]:
        """
        Add directory to ipfs.

        It wraps into directory.

        :param dir_path: str, path to dir to publish
        :param pin: bool, pin object or not
        :param recursive: bool, publish dierctory recursively or not
        :param wrap_with_directory: bool, wrap object with directory or not

        :return: dir name published, hash, list of items processed
        """
        response = self.client.add(
            dir_path,
            pin=pin,
            recursive=recursive,
            wrap_with_directory=wrap_with_directory,
        )
        if wrap_with_directory:
            return response[-2]["Name"], response[-1]["Hash"], response[:-1]

        if Path(dir_path).is_dir():
            return response[-1]["Name"], response[-1]["Hash"], response[:-1]

        return response["Name"], response["Hash"], []

    def pin(self, hash_id: str) -> Dict:
        """Pin content with hash_id"""

        try:
            return self.client.pin.add(hash_id, recursive=True)
        except ipfshttpclient.exceptions.ErrorResponse as e:
            raise PinError(f"Error on while pinning {hash_id}: {str(e)}") from e

    def remove(self, hash_id: str) -> Dict:
        """
        Remove dir added by it's hash.

        :param hash_id: str. hash of dir to remove

        :return: dict with unlinked items.
        """
        try:
            return self.client.pin.rm(hash_id, recursive=True)
        except ipfshttpclient.exceptions.ErrorResponse as e:
            raise RemoveError(f"Error on {hash_id} remove: {str(e)}") from e

    def remove_unpinned_files(self) -> None:
        """Remove dir added by it's hash."""
        try:
            return self.client.repo.gc()
        except ipfshttpclient.exceptions.ErrorResponse as e:
            raise RemoveError(
                f"Error while performing garbage collection: {str(e)}"
            ) from e

    def download(
        self,
        hash_id: str,
        target_dir: Union[str, Path],
        fix_path: bool = True,
        attempts: int = 5,
    ) -> str:
        """
        Download dir by its hash.

        :param hash_id: str. hash of file or package to download
        :param target_dir: Union[str, Path]. directory to place downloaded
        :param fix_path: bool. default True. on download don't wrap result in to hash_id directory.
        :param attempts: int. default 5. How often to attempt the download.
        :return: downloaded path
        """

        def move_to_target_dir(download_tmp_path: Path) -> str:
            """Move downloaded content to target directory"""
            if download_tmp_path.is_file():
                # Downloaded object is a file and can be moved easily
                shutil.move(str(download_tmp_path), str(target_dir))
                return str(target_dir)

            if fix_path:
                # identify name of package
                paths = list(download_tmp_path.glob("*"))
                if len(paths) == 1:
                    (package_path,) = paths
                else:
                    package_path = download_tmp_path

                # move content of directory with hashid to target dir
                shutil.move(str(package_path), str(target_dir))
                return str(Path(target_dir, package_path.name))

            # move directory with hash id to target dir
            shutil.move(str(download_tmp_path), str(target_dir))
            return str(Path(target_dir, download_tmp_path.name))

        target_dir = Path(target_dir)
        target_dir.mkdir(exist_ok=True, parents=True)

        while attempts:
            attempts -= 1
            try:  # download to tmp_dir in case of midway download failure
                with tempfile.TemporaryDirectory() as tmp_dir:
                    self.client.get(hash_id, tmp_dir)
                    return move_to_target_dir(Path(tmp_dir) / hash_id)
            except ipfshttpclient.exceptions.StatusError as e:
                logging.error(f"error on download of {hash_id}: {e}")
                time.sleep(1)

        raise DownloadError(f"Failed to download: {hash_id}")

    def publish(self, hash_id: str) -> Dict:
        """
        Publish directory by it's hash id.

        :param hash_id: hash of the directory to publish.

        :return: dict of names it was publish for.
        """
        try:
            return self.client.name.publish(hash_id)
        except ipfshttpclient.exceptions.TimeoutError as e:  # pragma: nocover
            raise PublishError(
                "can not publish within timeout, check internet connection!"
            ) from e

    def check_ipfs_node_running(self) -> None:
        """Check ipfs node running."""
        try:
            self.client.id()
        except ipfshttpclient.exceptions.CommunicationError as e:
            raise NodeError(f"Can not connect to node. Is node running?:\n{e}") from e



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test ipfs core."""

from unittest import mock

from aea_cli_ipfs.core import _get_path_data, register_package


def test_get_path_data() -> None:
    """Test Get path data."""
    with mock.patch(
        "aea_cli_ipfs.core.glob",
        return_value=["./packages/valory/connection/some/connection.yaml"],
    ):
        assert _get_path_data("./packages/") == (
            "./packages/valory/connection/some",
            "connection",
        )


def test_register_package() -> None:
    """Test register package."""
    with mock.patch(
        "aea_cli_ipfs.core.register_item_to_local_registry"
    ) as register_mock, mock.patch(
        "aea_cli_ipfs.core.glob",
        return_value=["./packages/valory/connection/some/connection.yaml"],
    ), mock.patch(
        "aea_cli_ipfs.core.load_item_config"
    ):
        ipfs_tool_mock = mock.Mock()
        ipfs_tool_mock.add = mock.Mock(
            return_value=(
                "some_name",
                "QmcRD4wkPPi6dig81r5sLj9Zm1gDCL4zgpEj9CfuRrGbzF",
                None,
            )
        )

        assert (
            register_package(ipfs_tool_mock, "some_dir", no_pin=True)
            == "bafybeigrf2dwtpjkiovnigysyto3d55opf6qkdikx6d65onrqnfzwgdkfa"
        )

        register_mock.assert_called_once_with(
            item_type="connection",
            public_id=mock.ANY,
            package_hash="bafybeigrf2dwtpjkiovnigysyto3d55opf6qkdikx6d65onrqnfzwgdkfa",
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test test_tools."""

from unittest.mock import patch

import pytest
from _pytest.compat import get_real_func
from aea_cli_ipfs.test_tools.fixture_helpers import ipfs_daemon


def test_daemon_fixture() -> None:
    """Test ipfs daemon fixture."""
    with patch("aea_cli_ipfs.ipfs_utils.IPFSDaemon.start") as start_mock, patch(
        "aea_cli_ipfs.ipfs_utils.IPFSDaemon.stop"
    ) as stop_mock:
        gen = get_real_func(ipfs_daemon)()

        assert next(gen) is False
        with pytest.raises(StopIteration):
            next(gen)
        start_mock.assert_called_once()
        stop_mock.assert_called_once()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Core components for `ipfs cli command`."""
import os
from glob import glob
from pathlib import Path
from typing import Any, Optional, Tuple

import click
from aea_cli_ipfs.ipfs_utils import (
    DownloadError,
    IPFSTool,
    NodeError,
    PublishError,
    RemoveError,
)
from aea_cli_ipfs.registry import register_item_to_local_registry

from aea.cli.utils.config import get_ipfs_node_multiaddr, load_item_config
from aea.configurations.constants import CONFIG_FILE_TO_PACKAGE_TYPE
from aea.helpers.cid import to_v1


@click.group()
@click.pass_context
def ipfs(click_context: click.Context) -> None:
    """IPFS Commands"""
    ipfs_tool = IPFSTool(get_ipfs_node_multiaddr())
    click_context.obj = ipfs_tool
    try:
        ipfs_tool.check_ipfs_node_running()
    except NodeError:
        click.echo("Can not connect to the local ipfs node. Starting own one.")
        ipfs_tool.daemon.start()


@ipfs.resultcallback()
@click.pass_context
def process_result(click_context: click.Context, *_: Any, **__: Any) -> None:
    """Tear down command group."""
    ipfs_tool = click_context.obj
    if ipfs_tool.daemon.is_started_internally():  # pragma: nocover
        click.echo("Stopping ipfs node launched to execute the command.")
        ipfs_tool.daemon.stop()
        click.echo("Daemon stopped.")


@ipfs.command()
@click.argument(
    "dir_path",
    type=click.Path(
        exists=True, dir_okay=True, file_okay=False, resolve_path=True, readable=True
    ),
    required=False,
)
@click.option("-p", "--publish", is_flag=True)
@click.option("--no-pin", is_flag=True)
@click.pass_context
def add(
    click_context: click.Context,
    dir_path: Optional[str],
    publish: bool = False,
    no_pin: bool = False,
) -> None:
    """Add directory to ipfs, if not directory specified the current one will be added."""
    dir_path = dir_path or os.getcwd()
    ipfs_tool = click_context.obj
    package_hash = register_package(ipfs_tool, dir_path, no_pin)
    if publish:
        click.echo("Publishing...")
        try:
            response = ipfs_tool.publish(package_hash)
            click.echo(f"Published to {response['Name']}")
        except PublishError as e:
            raise click.ClickException(f"Publish failed: {str(e)}") from e


@ipfs.command()
@click.argument(
    "hash_",
    metavar="hash",
    type=str,
    required=True,
)
@click.pass_context
def remove(click_context: click.Context, hash_: str) -> None:
    """Remove a directory from ipfs by it's hash."""
    ipfs_tool = click_context.obj
    try:
        ipfs_tool.remove(hash_)
        click.echo(f"{hash_} was removed successfully")
    except RemoveError as e:
        raise click.ClickException(f"Remove error: {str(e)}") from e


@ipfs.command()
@click.argument(
    "hash_",
    metavar="hash",
    type=str,
    required=True,
)
@click.argument(
    "target_dir",
    type=click.Path(dir_okay=True, file_okay=False, resolve_path=True),
    required=False,
)
@click.pass_context
def download(
    click_context: click.Context, hash_: str, target_dir: Optional[str]
) -> None:
    """Download directory by it's hash, if not target directory specified will use current one."""
    target_dir = target_dir or os.getcwd()
    ipfs_tool = click_context.obj
    click.echo(f"Download {hash_} to {target_dir}")
    try:
        ipfs_tool.download(hash_, target_dir)
        click.echo("Download complete!")
    except DownloadError as e:  # pragma: nocover
        raise click.ClickException(str(e)) from e


def _get_path_data(dir_path: str) -> Optional[Tuple[str, str]]:
    """
    Returns the file path for item config file.

    :param dir_path: directory path.
    :return: package path and item type.
    """

    yaml_files = glob(str(Path(dir_path) / "*.yaml"))
    for config_file_path in yaml_files:
        package_path, config_file = os.path.split(config_file_path)
        if config_file in CONFIG_FILE_TO_PACKAGE_TYPE.keys():
            return (package_path, CONFIG_FILE_TO_PACKAGE_TYPE[config_file])
    return None


def register_package(
    ipfs_tool: IPFSTool,
    dir_path: str,
    no_pin: bool,
) -> str:
    """
    Register package to IPFS registry.

    :param ipfs_tool: instance of IPFSTool.
    :param dir_path: package directory.
    :param no_pin: pin object or not.
    :return: package hash
    """

    click.echo(f"Processing package: {dir_path}")
    name, package_hash, _ = ipfs_tool.add(dir_path, pin=(not no_pin))
    package_hash = to_v1(package_hash)
    path_data = _get_path_data(dir_path)
    if path_data is not None:
        package_path, item_type = path_data
        package_path_ = Path(package_path)
        item_config = load_item_config(item_type=item_type, package_path=package_path_)
        register_item_to_local_registry(
            item_type=item_type,
            public_id=item_config.public_id,
            package_hash=package_hash,
        )
        click.echo(
            f"Registered item with:\n\titem_type: {item_type}\n\tpublic id : {item_config.public_id}\n\thash : {package_hash}"
        )
    else:
        click.echo(f"Added: `{name}`, hash is {package_hash}")
    return package_hash



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------


"""Setup script for the plug-in."""


from setuptools import setup  # type: ignore


setup(
    name="open-aea-cli-ipfs",
    version="1.39.0.post1",
    author="Valory AG",
    license="Apache-2.0",
    description="CLI extension for open AEA framework wrapping IPFS functionality.",
    long_description="CLI extension for open AEA framework wrapping IPFS functionality.",
    long_description_content_type="text/markdown",
    packages=["aea_cli_ipfs"],
    entry_points={"aea.cli": ["ipfs_cli_command = aea_cli_ipfs.core:ipfs"]},
    install_requires=[
        "open-aea>=1.0.0, <2.0.0",
        "ipfshttpclient>=0.8.0a2",
        "pytest>=7.0.0,<7.3.0",
    ],
    classifiers=[
        "Environment :: Console",
        "Environment :: Web Environment",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Communications",
        "Topic :: Internet",
        "Topic :: Software Development",
    ],
)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Module with methods for ipfs registry."""

import json
import logging
import os
from pathlib import Path
from typing import Dict, List, Optional, Union

import jsonschema
from aea_cli_ipfs.exceptions import HashNotProvided
from aea_cli_ipfs.ipfs_utils import DownloadError, IPFSTool, NodeError

from aea.cli.registry.settings import DEFAULT_IPFS_URL_LOCAL
from aea.cli.utils.config import get_ipfs_node_multiaddr
from aea.configurations.base import PublicId


_default_logger = logging.getLogger(__name__)

LocalRegistry = Dict[str, Dict[str, str]]

LOCAL_REGISTRY_PATH = os.path.join(
    os.path.expanduser("~"), ".aea", "local_registry.json"
)
LOCAL_REGISTRY_DEFAULT: LocalRegistry = {
    "protocols": {},
    "skills": {},
    "connections": {},
    "contracts": {},
    "agents": {},
}

LOCAL_REGISTRY_SCHEMA = {
    "type": "object",
    "properties": {
        "protocols": {
            "type": "object",
            "propertyNames": {"pattern": r"^[a-z][a-z0-9_]+\/[a-z_0-9]+:\d\.\d\.\d$"},
        },
        "skills": {"type": "object"},
        "connections": {"type": "object"},
        "contracts": {"type": "object"},
        "agents": {"type": "object"},
    },
    "required": ["protocols", "skills", "connections", "contracts", "agents"],
}


def validate_registry(registry_data: LocalRegistry) -> None:
    """
    Validate local registry data.

    :param registry_data: json like object containing registry data.
    """
    try:
        jsonschema.validate(registry_data, schema=LOCAL_REGISTRY_SCHEMA)
    except jsonschema.ValidationError as e:
        _default_logger.debug("Registry Not Valid")
        raise ValueError(str(e))


def write_local_registry(
    registry_data: LocalRegistry, registry_path: str = LOCAL_REGISTRY_PATH
) -> None:
    """
    Write registry data to file.

    :param registry_data: json like object containing registry data.
    :param registry_path: local registry path.
    """
    validate_registry(registry_data)
    with open(registry_path, mode="w+", encoding="utf-8") as fp:
        json.dump(registry_data, fp)


def load_local_registry(registry_path: str = LOCAL_REGISTRY_PATH) -> LocalRegistry:
    """Returns local registry data."""

    local_registry_path = Path(registry_path)
    if not local_registry_path.is_file():
        write_local_registry(LOCAL_REGISTRY_DEFAULT, registry_path=registry_path)
        return LOCAL_REGISTRY_DEFAULT

    with open(local_registry_path, mode="r", encoding="utf-8") as fp:
        registry_data = json.load(fp)
        validate_registry(registry_data)
        return registry_data


def get_ipfs_hash_from_public_id(
    item_type: str,
    public_id: PublicId,
    registry_path: str = LOCAL_REGISTRY_PATH,
) -> Optional[str]:
    """Get IPFS hash from local registry."""

    registry_data = load_local_registry(registry_path=registry_path)
    if public_id.package_version.is_latest:
        package_versions: List[PublicId] = [
            PublicId.from_str(_public_id)
            for _public_id in registry_data.get(f"{item_type}s", {}).keys()
            if public_id.same_prefix(PublicId.from_str(_public_id))
        ]
        package_versions = list(
            reversed(sorted(package_versions, key=lambda x: x.package_version))
        )
        if len(package_versions) == 0:
            return None
        public_id, *_ = package_versions

    return registry_data.get(f"{item_type}s", {}).get(str(public_id), None)


def register_item_to_local_registry(
    item_type: str,
    public_id: Union[str, PublicId],
    package_hash: str,
    registry_path: str = LOCAL_REGISTRY_PATH,
) -> None:
    """
    Add PublicId to hash mapping in the local registry.

    :param item_type: item type.
    :param public_id: public id of package.
    :param package_hash: hash of package.
    :param registry_path: local registry path.
    """

    registry_data = load_local_registry(registry_path=registry_path)
    registry_data[f"{item_type}s"][str(public_id)] = str(package_hash)
    write_local_registry(registry_data, registry_path)


def fetch_ipfs(
    item_type: str,
    public_id: PublicId,
    dest: str,
    remote: bool = True,
) -> Optional[Path]:
    """Fetch a package from IPFS node."""
    if remote:
        ipfs_tool = IPFSTool(get_ipfs_node_multiaddr())
    else:
        ipfs_tool = IPFSTool(addr=DEFAULT_IPFS_URL_LOCAL)

    try:
        package_hash: Optional[str] = public_id.hash
    except ValueError:
        package_hash = (
            None if remote else get_ipfs_hash_from_public_id(item_type, public_id)
        )

    if package_hash is None:
        raise HashNotProvided(f"Please provide hash; Public id {public_id}.")

    try:
        ipfs_tool.check_ipfs_node_running()
    except NodeError:  # pragma: nocover
        if not remote:
            ipfs_tool.daemon.start()
        else:
            raise Exception(f"Cannot connect to node with addr: {ipfs_tool.addr}")

    try:
        *_download_dir, _ = os.path.split(dest)
        download_dir = os.path.sep.join(_download_dir)
        ipfs_tool.download(package_hash, download_dir)
        package_path = Path(dest).absolute()
        ipfs_tool.daemon.stop()
        return package_path

    except DownloadError as e:  # pragma: nocover
        ipfs_tool.daemon.stop()
        raise Exception(str(e)) from e
